@@ -176,7 +176,8 @@ public:
     auto stopped = false;
 
     mIo->async([this, &mutex, &condition, &stopped]() {
-      enable(false);
+      mEnabled = false;
+      mDiscovery.enable(false);
       std::unique_lock<std::mutex> lock(mutex);
       stopped = true;
       condition.notify_one();
@@ -193,18 +194,7 @@ public:
     const bool bWasEnabled = mEnabled.exchange(bEnable);
     if (bWasEnabled != bEnable)
     {
-      mIo->async([this, bEnable] {
-        if (bEnable)
-        {
-          // Process the pending client states to make sure we don't push one after we
-          // have joined a running session
-          mRtClientStateSetter.processPendingClientStates();
-          // Always reset when first enabling to avoid hijacking
-          // tempo in existing sessions
-          resetState();
-        }
-        mDiscovery.enable(bEnable);
-      });
+      mRtClientStateSetter.invoke();
     }
   }
 
@@ -570,11 +560,23 @@ private:
     RtClientStateSetter(Controller& controller)
       : mController(controller)
       , mCallbackDispatcher(
-          [this] { mController.mIo->async([this]() { processPendingClientStates(); }); },
+          [this] {
+            mController.mIo->async([this]() {
+              // Process the pending client states first to make sure we don't push one
+              // after we have joined a running session when enabling
+              processPendingClientStates();
+              updateEnabled();
+            });
+          },
           detail::kRtHandlerFallbackPeriod)
     {
     }
 
+    void invoke()
+    {
+      mCallbackDispatcher.invoke();
+    }
+
     void push(const IncomingClientState clientState)
     {
       if (clientState.timeline)
@@ -600,6 +602,20 @@ private:
       mController.handleRtClientState(clientState);
     }
 
+    void updateEnabled()
+    {
+      if (mController.mEnabled && !mController.mDiscovery.isEnabled())
+      {
+        // Always reset when first enabling to avoid hijacking tempo in existing sessions
+        mController.resetState();
+        mController.mDiscovery.enable(true);
+      }
+      else if (!mController.mEnabled && mController.mDiscovery.isEnabled())
+      {
+        mController.mDiscovery.enable(false);
+      }
+    }
+
   private:
     IncomingClientState buildMergedPendingClientState()
     {
@@ -717,18 +733,11 @@ private:
   {
     GatewayPtr operator()(std::pair<NodeState, GhostXForm> state,
       util::Injected<IoType&> io,
-      const asio::ip::address& addr)
+      const discovery::IpAddress& addr)
     {
-      if (addr.is_v4())
-      {
-        return GatewayPtr{new ControllerGateway{std::move(io), addr.to_v4(),
-          util::injectVal(makeGatewayObserver(mController.mPeers, addr)),
-          std::move(state.first), std::move(state.second), mController.mClock}};
-      }
-      else
-      {
-        throw std::runtime_error("Could not create peer gateway on non-ipV4 address");
-      }
+      return GatewayPtr{new ControllerGateway{std::move(io), addr,
+        util::injectVal(makeGatewayObserver(mController.mPeers, addr)),
+        std::move(state.first), std::move(state.second), mController.mClock}};
     }
 
     Controller& mController;
